#include <stdio.h>
#include <string.h>
#include "cv.h"
#include "cxcore.h"
#include "highgui.h"

#define MAXLABEL 65535

#define PIXVAL(iplimagep, x, y)  (*(uchar *)((iplimagep)->imageData  + (y) * (iplimagep)->widthStep + (x)))
#define PIXVAL_16U(iplimagep, x, y)  (*(ushort *)((iplimagep)->imageData  + (y) * (iplimagep)->widthStep + (2 * (x))))

int pval(IplImage *label, int x, int y)
{
    if (x < 0 || y < 0 || x >= label->width || y >= label->height) {
        return 0;
    }
    return PIXVAL_16U(label, x, y);
}

void updateConnection(ushort *con, ushort l1, ushort l2, int curlabel)
{
    // XXX: can be faster
    
    int i;
    ushort p1 = con[l1]; // parent
    ushort p2 = con[l2]; // parent

    if (p1 == p2) {
        return;
    }
    if (p1 > p2) {
        ushort t = p1;
        p1 = p2;
        p2 = t;
    }
    for (i = 0; i <= curlabel; i++) {
        if (con[i] == p2) {
            con[i] = p1;
        }
    }
}

int
main()
{
    IplImage *img, *work, *label, *rslt;
    CvFont font;
    int i, j;
    ushort curlabel;
    ushort u_label, l_label;
    ushort connect[MAXLABEL + 1];
    int m00[MAXLABEL + 1];
    int m10[MAXLABEL + 1];
    int m01[MAXLABEL + 1];
    char str[256];

    CvCapture *capture = cvCaptureFromCAM(0);

    if (capture == NULL) {
        fprintf(stderr, "cannot capture\n");
        return 1;
    }

    img = cvQueryFrame(capture);
    work = cvCreateImage(cvGetSize(img), IPL_DEPTH_8U, 1);
    rslt = cvCreateImage(cvGetSize(img), IPL_DEPTH_8U, 3);
    label = cvCreateImage(cvGetSize(img), IPL_DEPTH_16U, 1);

    cvNamedWindow("image", CV_WINDOW_AUTOSIZE);
    cvNamedWindow("rslt", CV_WINDOW_AUTOSIZE);

    while (1) {
        img = cvQueryFrame(capture); /* cannot be modified */
        cvCvtColor(img, work, CV_BGR2GRAY);
        cvFlip(work, work, NULL);

        for (j = 0; j < work->height; j++) {
            for (i = 0; i < work->width; i++) {
                PIXVAL(work, i, j) = (PIXVAL(work, i, j) > 150)? 255: 0;
            }
        }
        cvErode(work, work, NULL, 3);
        cvDilate(work, work, NULL, 3);

        cvSet(label, cvScalar(0));
        curlabel = 0;
        for (i = 0; i <= MAXLABEL; i++) {
            connect[i] = 0;
            m00[i] = 0;
            m10[i] = 0;
            m01[i] = 0;
        }
        for (j = 0; j < work->height; j++) {
            for (i = 0; i < work->width; i++) {
                if (PIXVAL(work, i, j) != 0) {
                    u_label = pval(label, i, j - 1);
                    l_label = pval(label, i - 1, j);
                    if (u_label > 0 && l_label > 0) {
                        if (u_label == l_label) {
                            PIXVAL_16U(label, i, j) = u_label;
                        } else if (u_label < l_label) {
                            PIXVAL_16U(label, i, j) = u_label;
                            updateConnection(connect, u_label, l_label,
                                             curlabel);
                        } else {
                            PIXVAL_16U(label, i, j) = l_label;
                            updateConnection(connect, l_label, u_label,
                                             curlabel);
                        }
                    } else if (u_label > 0 && l_label == 0) {
                        PIXVAL_16U(label, i, j) = u_label;
                    } else if (u_label == 0 && l_label > 0) {
                        PIXVAL_16U(label, i, j) = l_label;
                    } else if (curlabel < MAXLABEL) {
                        PIXVAL_16U(label, i, j) = ++curlabel;
                        connect[curlabel] = curlabel;
                    }
                }
            }
        }

        {
            int last_update = 0;
            int cur_real_label = 0;
            for (i = 1; i <= curlabel; i++) {
                if (connect[i] > last_update) {
                    last_update = connect[i];
                    connect[i] = ++cur_real_label;
                } else {
                    connect[i] = connect[connect[i]];
                }
            }
        }

        for (j = 0; j < work->height; j++) {
            for (i = 0; i < work->width; i++) {
                ushort l = connect[PIXVAL_16U(label, i, j)];
                PIXVAL_16U(label, i, j) = l;
                m00[l]++;
                m10[l] += i;
                m01[l] += j;
            }
        }

        cvCvtColor(work, rslt, CV_GRAY2BGR);
        cvInitFont(&font, CV_FONT_HERSHEY_PLAIN, 1.0, 1.0);
        for (i = 1; i <= curlabel; i++) {
            int gx, gy;
            if (m00[i] == 0) {
                continue;
            }
            gx = m10[i] / m00[i];
            gy = m01[i] / m00[i];
            sprintf(str, "%d", i);
            cvPutText(rslt, str, cvPoint(gx, gy), &font,
                      CV_RGB(255, 0, 0));
        }
    
        cvShowImage("image", img);
        cvShowImage("rslt", rslt);

        if (cvWaitKey(30) > 0) {
            break;
        }
    }

    return 0;
}
