#include <stdio.h>
#include "cv.h"
#include "cxcore.h"
#include "highgui.h"
#include "perftimer.h"

//#define OPT_NAIVE
//#define OPT_INTERCHANGE
#define OPT_INDEX

#define PIXVAL(iplimagep, x, y) (*(uchar *)((iplimagep)->imageData + (y) * (iplimagep)->widthStep + (x)))
#define PIXVALC(iplimagep, x, y, c) (*(uchar *)((iplimagep)->imageData + (y) * (iplimagep)->widthStep + (3 * (x)) + c))

perftimer *pt;

int main()
{
    int i, j;
    int m, d;
    uchar r, g, b;
    int width, height;
    int wpad1, wpad3;
    uchar *cptr;
    uchar *mptr;
    uchar *bptr;
    int loop = 200;

    CvCapture *capture = cvCaptureFromCAM(0);

    IplImage *frame, *color, *mono, *back;
    pt = create_perftimer(loop, 1);

    if (!capture) {
	fprintf(stderr, "cannot capture\n");
	return 1;
    }
    frame = cvQueryFrame(capture);
    color = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 3);  // color
    mono = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 1);
    back = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 1);

    width = color->width;
    height = color->height;
    wpad1 = mono->widthStep - width;
    wpad3 = color->widthStep - 3 * width;

    cvNamedWindow("image", CV_WINDOW_AUTOSIZE);

    while (1) {
	frame = cvQueryFrame(capture); /* cannot be modified */
	cvFlip(frame, color, 0);
	//cvCvtColor(color, mono, CV_BGR2GRAY);
	start_perftimer(pt, 0);

#if defined(OPT_NAIVE)
	//cvCvtColor(color, mono, CV_BGR2GRAY);
	for (j = 0; j < height; j++) {
	    for (i = 0; i < width; i++) {
		b = PIXVALC(color, i, j, 0);
		g = PIXVALC(color, i, j, 1);
		r = PIXVALC(color, i, j, 2);
		PIXVAL(mono, i, j) = (r * 306 + g * 601 + b * 117) / 1024;
	    }
	}

	// frame diff
	for (j = 0; j < height; j++) {
	    for (i = 0; i < width; i++) {
		m = PIXVAL(mono, i, j);
		d = m - PIXVAL(back, i, j);
		if (d > 30 || d < -30) {
		    r = 0;
		    g = 255;
		    b = 0;
		    PIXVALC(color, i, j, 0) = b;
		    PIXVALC(color, i, j, 1) = g;
		    PIXVALC(color, i, j, 2) = r;
		} else {
		    PIXVALC(color, i, j, 0) = m;
		    PIXVALC(color, i, j, 1) = m;
		    PIXVALC(color, i, j, 2) = m;
		}
	    }
	}
	    
	//cvCopy(mono, back, 0);
	for (j = 0; j < height; j++) {
	    for (i = 0; i < width; i++) {
		PIXVAL(back, i, j) = PIXVAL(mono, i, j);
	    }
	}
#elif defined(OPT_INTERCHANGE)
	//cvCvtColor(color, mono, CV_BGR2GRAY);
	for (i = 0; i < width; i++) {
	    for (j = 0; j < height; j++) {
		b = PIXVALC(color, i, j, 0);
		g = PIXVALC(color, i, j, 1);
		r = PIXVALC(color, i, j, 2);
		PIXVAL(mono, i, j) = (r * 306 + g * 601 + b * 117) / 1024;
	    }
	}

	// frame diff
	for (i = 0; i < width; i++) {
	    for (j = 0; j < height; j++) {
		m = PIXVAL(mono, i, j);
		d = m - PIXVAL(back, i, j);
		if (d > 30 || d < -30) {
		    r = 0;
		    g = 255;
		    b = 0;
		    PIXVALC(color, i, j, 0) = b;
		    PIXVALC(color, i, j, 1) = g;
		    PIXVALC(color, i, j, 2) = r;
		} else {
		    PIXVALC(color, i, j, 0) = m;
		    PIXVALC(color, i, j, 1) = m;
		    PIXVALC(color, i, j, 2) = m;
		}
	    }
	}
	    
	//cvCopy(mono, back, 0);
	for (i = 0; i < width; i++) {
	    for (j = 0; j < height; j++) {
		PIXVAL(back, i, j) = PIXVAL(mono, i, j);
	    }
	}
#elif defined(OPT_INDEX)
	//cvCvtColor(color, mono, CV_BGR2GRAY);
	cptr = (uchar *)(color->imageData);
	mptr = (uchar *)(mono->imageData); 
	for (j = 0;
	     j < height;
	     j++, cptr += wpad3, mptr += wpad1) {
	    for (i = 0;
		 i < width;
		 i++, cptr += 3, mptr++) {
		b = *cptr;
		g = *(cptr + 1);
		r = *(cptr + 2);
		*mptr = (r * 306 + g * 601 + b * 117) / 1024;
	    }
	}

	// frame diff
	cptr = (uchar *)(color->imageData);
	mptr = (uchar *)(mono->imageData);
	bptr = (uchar *)(back->imageData);
	for (j = 0;
	     j < height;
	     j++, cptr += wpad3, mptr += wpad1, bptr += wpad1) {
	    for (i = 0;
		 i < width;
		 i++, cptr += 3, mptr++, bptr++) {
		m = *mptr;
		d = m - (*bptr);
		if (d > 30 || d < -30) {
		    r = 0;
		    g = 255;
		    b = 0;
		    *cptr = b;
		    *(cptr + 1) = g;
		    *(cptr + 2) = r;
		} else {
		    *cptr = m;
		    *(cptr + 1) = m;
		    *(cptr + 2) = m;
		}
	    }
	}

	//cvCopy(mono, back, 0);
	mptr = (uchar *)(mono->imageData);
	bptr = (uchar *)(back->imageData);
	for (j = 0;
	     j < height;
	     j++, mptr += wpad1, bptr += wpad1) {
	    for (i = 0;
		 i < width;
		 i++, mptr++, bptr++) {
		*bptr = *mptr;
	    }
	}
#endif

	stop_perftimer(pt, 0);

	cvShowImage("image", color);
	//cvShowImage("image", mono);

	if (cvWaitKey(30) > 0) {
	    break;
	}
    }

    print_stat_perftimer(pt, stdout);

    return 0;
}
